# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from abc import abstractmethod
from dataclasses import asdict, dataclass
from typing import Optional, Type, TypeVar

import torch

"""
Feature maps allow for a given query or key to be encoded in a different space.
"""

Self = TypeVar("Self", bound="FeatureMap")


@dataclass
class FeatureMapConfig:
    name: str
    dim_features: int
    iter_before_redraw: Optional[int]
    normalize_inputs: Optional[bool]
    epsilon: Optional[float]


class FeatureMap(torch.nn.Module):
    def __init__(
        self,
        dim_features: int,
        iter_before_redraw: Optional[int] = None,
        normalize_inputs: bool = False,
        epsilon: float = 1e-6,
    ):
        super().__init__()

        self.dim_features = dim_features
        self.dim_feature_map = dim_features

        self.iter_before_redraw = iter_before_redraw
        self.features: Optional[torch.Tensor] = None
        self.epsilon = epsilon
        self.normalize_inputs = normalize_inputs

        self._iter_counter = 0

    @abstractmethod
    def _get_feature_map(self, dim_input: int, dim_features: int, device: torch.device):
        raise NotImplementedError()

    @classmethod
    def from_config(cls: Type[Self], config: FeatureMapConfig) -> Self:
        # Generate the class inputs from the config
        fields = asdict(config)

        # Skip all Nones so that default values are used
        fields = {k: v for k, v in fields.items() if v is not None}

        return cls(**fields)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import tempfile

import torch


def init_torch_distributed_local():
    if torch.distributed.is_initialized():
        return

    init_url = "file://" + tempfile.mkstemp()[1]
    backend = (
        torch.distributed.Backend.NCCL
        if torch.cuda.is_available()
        else torch.distributed.Backend.GLOO
    )
    torch.distributed.init_process_group(
        backend=backend,
        rank=0,
        world_size=1,
        init_method=init_url,
    )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import itertools
import math
from functools import partial
from typing import cast

import torch
from torch.utils import benchmark
from utils import benchmark_main_helper

import xformers.ops

torch.backends.cuda.matmul.allow_tf32 = False


def create_attn_bias(
    bias_type, batch_size: int, q_len: int, kv_len: int, device, dtype
):
    NoneType = type(None)
    if bias_type is NoneType:
        return None
    if bias_type is torch.Tensor:
        attn_bias = torch.randn((batch_size, 1, kv_len), device=device, dtype=dtype) * 3
        return attn_bias.expand(batch_size, q_len, kv_len)
    if bias_type is xformers.ops.LowerTriangularMask:
        return bias_type([batch_size, q_len, kv_len], dtype=dtype, device=device)
    assert False, f"Unsupported bias type: {bias_type}"


def ref_attention(q, k, v, attn_bias=None, p=0.0):
    if isinstance(attn_bias, xformers.ops.AttentionMask):
        attn_bias = attn_bias.to_tensor().to(q.dtype)
    q = q * (1.0 / q.shape[-1] ** 0.5)
    if attn_bias is None:
        attn = q @ k.transpose(-2, -1)
    else:
        # equivalent to (q @ k.transpose(-2, -1) + m).softmax(-1) @ v
        # but faster, and is what is used in PyTorch now
        attn = torch.baddbmm(attn_bias, q, k.transpose(-2, -1))
    attn = attn.softmax(-1)
    if p > 0:
        attn = torch.nn.functional.dropout(attn, p=p)
    return attn @ v


min_run_time = 2
device = torch.device("cuda")

NUM_THREADS = [1] if device.type == "cuda" else [1, 40]
SHAPES = list(itertools.product([32, 256], [128, 512, 1024], [16, 32, 64, 128]))
SHAPES = list(set(SHAPES))
SHAPES.sort()


p = 0.0
FORCE_OP = None
# FORCE_OP = xformers.ops.MemoryEfficientAttentionOp
# FORCE_OP = xformers.ops.MemoryEfficientAttentionGenericForwardOp
# FORCE_OP = xformers.ops.MemoryEfficientAttentionFlashAttentionOp


def product_dict(**kwargs):
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in itertools.product(*vals):
        yield dict(zip(keys, instance))


CASES = list(
    product_dict(
        shape=SHAPES,
        num_threads=NUM_THREADS,
        attn_bias_type=[type(None), torch.Tensor, xformers.ops.LowerTriangularMask],
        dtype=[torch.half, torch.bfloat16, torch.float],
    )
)


def benchmark_forward(shape, num_threads: int, attn_bias_type, dtype):
    B, M, K = shape
    q = torch.rand(shape, device=device, dtype=dtype)

    dispatch = xformers.ops.AttentionOpDispatch(
        dtype=dtype,
        device=device,
        k=K,
        attn_bias_type=attn_bias_type,
        has_dropout=False,
        kv_len=M,
        q_len=M,
    )
    try:
        op = dispatch.op if FORCE_OP is None else FORCE_OP
    except NotImplementedError:
        return
    if not op.supports(dispatch):
        return

    attn_bias = create_attn_bias(
        attn_bias_type,
        batch_size=B,
        q_len=M,
        kv_len=M,
        device=device,
        dtype=dtype,
    )

    dtype_str = {
        torch.bfloat16: "b16",
        torch.half: "f16",
        torch.float: "f32",
    }[dtype]
    sub_label = f"{dtype_str} {op.NAME} B={B}, M={M}, K={K}"

    if True:
        r = xformers.ops.memory_efficient_attention(q, q, q, attn_bias, op=op).float()
        rr = ref_attention(
            q.float(),
            q.float(),
            q.float(),
            attn_bias,
        )
        assert (r - rr).abs().max() < 4e-3, (r - rr).abs().max()
        del r, rr

    yield benchmark.Timer(
        stmt="fn(q, q, q, attn_bias, p)",
        globals={
            "q": q,
            "attn_bias": attn_bias,
            "p": p,
            "fn": partial(xformers.ops.memory_efficient_attention, op=op),
        },
        label=f"attention (attn_bias={attn_bias_type})",
        description="optimized",
        sub_label=sub_label,
        num_threads=num_threads,
    )
    yield benchmark.Timer(
        stmt="fn(q, q, q, attn_bias, p)",
        globals={
            "q": q,
            "attn_bias": attn_bias,
            "p": p,
            "fn": ref_attention,
        },
        label=f"attention (attn_bias={attn_bias_type})",
        description="vanilla",
        sub_label=sub_label,
        num_threads=num_threads,
    )


def benchmark_backward(shape, num_threads: int, attn_bias_type, dtype):
    B, M, K = shape
    q = torch.rand(shape, device=device, dtype=dtype, requires_grad=True)

    dispatch = xformers.ops.AttentionOpDispatch(
        dtype=dtype,
        device=device,
        k=K,
        attn_bias_type=attn_bias_type,
        has_dropout=False,
        kv_len=M,
        q_len=M,
    )
    try:
        op = dispatch.op if FORCE_OP is None else FORCE_OP
    except NotImplementedError:
        return
    if not op.supports(dispatch):
        return

    attn_bias = create_attn_bias(
        attn_bias_type,
        batch_size=B,
        q_len=M,
        kv_len=M,
        device=device,
        dtype=dtype,
    )

    dtype_str = {
        torch.bfloat16: "b16",
        torch.half: "f16",
        torch.float: "f32",
    }[dtype]
    sub_label = f"{dtype_str} {op.NAME} B={B}, M={M}, K={K}"

    if True:
        r = xformers.ops.memory_efficient_attention(q, q, q, attn_bias, op=op)
        r.backward(torch.ones_like(q))

        grad = cast(torch.Tensor, q.grad)
        q.grad = None

        rr = ref_attention(q, q, q, attn_bias)
        rr.backward(torch.ones_like(q))
        atol = 2e-4 + 2e-6 * K * M * math.sqrt(B) * math.sqrt(M)
        # type: ignore
        assert (grad - q.grad).abs().max() < atol, f"{(grad - q.grad).abs().max()}"
        q.grad = None
        del r, rr, grad

    out = xformers.ops.memory_efficient_attention(q, q, q, attn_bias, p, op=op)
    grad = torch.ones_like(q)

    yield benchmark.Timer(
        stmt="out.backward(grad, retain_graph=True)",
        globals={
            "out": out,
            "grad": grad,
        },
        label=f"attention backward (attn_bias={attn_bias_type})",
        description="optimized",
        sub_label=sub_label,
        num_threads=num_threads,
    )
    del out

    yield benchmark.Timer(
        stmt="out.backward(grad, retain_graph=True)",
        globals={
            "out": ref_attention(q, q, q, attn_bias, p),
            "grad": grad,
        },
        label=f"attention backward (attn_bias={attn_bias_type})",
        description="vanilla",
        sub_label=sub_label,
        num_threads=num_threads,
    )


benchmark_main_helper(benchmark_forward, CASES, min_run_time=min_run_time)
benchmark_main_helper(benchmark_backward, CASES, min_run_time=min_run_time)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import gc
import math
from collections import namedtuple
from dataclasses import dataclass

import matplotlib.pyplot as plt
import torch
import triton
from triton.ops.blocksparse import matmul as blocksparse_matmul

from xformers.benchmarks.utils import pretty_barplot
from xformers.components.attention.attention_patterns import (
    axial_2d_pattern,
    causal_1d_pattern,
    global_token_pattern,
    local_1d_pattern,
    local_2d_pattern,
)
from xformers.components.attention.core import SparseCS, _matmul_with_mask

device = "cuda"
TestCase = namedtuple("TestCase", ["prepare_callable", "mask", "config", "name"])

##############################################
# Plotting utilities
##############################################


def plot_mask(mask, config, filename):
    sparsity = get_sparsity(mask)
    batch_size = config.batch_size
    num_heads = config.num_heads
    seq_len = config.seq_length

    proxy = torch.ones(batch_size, num_heads, seq_len, seq_len, dtype=torch.bool)
    proxy = triton.testing.mask_tensor(proxy, mask, config.block_size, False)
    proxy = proxy[0][0]

    f = plt.figure()
    plt.imshow(proxy.logical_not(), cmap="gray")
    plt.suptitle("Sparsity = " + str(sparsity) + "%")
    plt.savefig(filename)
    plt.close(f)


##############################################
# Mask and testing utilities
##############################################


def get_mask(MaskGenType, config, config_setter=[]):
    mask_config = Configuration()
    mask_config.init(config)

    # Get the mask
    mask_generator = MaskGenType(mask_config)
    for (key, value) in config_setter:
        mask_generator.set_config_attr(key, value)
    if not mask_generator.is_valid_config():
        return None

    return mask_generator()


def densify_mask(mask, config):
    num_heads = config.num_heads
    seq_length = config.seq_length
    block_size = config.block_size
    dense_mask = torch.zeros(num_heads, seq_length, seq_length)
    for (h, i, j) in zip(*mask.nonzero(as_tuple=True)):
        dense_mask[
            h,
            i * block_size : (i + 1) * block_size,
            j * block_size : (j + 1) * block_size,
        ] = mask[h, i, j]
    return dense_mask


def mask_tensor(a, mask, config):
    return triton.testing.mask_tensor(a, mask, config.block_size, 0.0)


def sparsify_tensor(a, mask, config):
    return triton.testing.sparsify_tensor(a, mask, config.block_size)


def get_sparsity(mask):
    return round((1.0 - mask.sum().item() / mask.numel()) * 100)


##############################################
# Mask Generation
##############################################


@dataclass
class Configuration(object):
    batch_size: int = 32
    num_heads: int = 12
    seq_length: int = 2048
    hidden_size: int = 768  # hidden_size = n_heads * projection_hidden_dimension
    block_size: int = 64

    @property
    def blocked_seq_length(self):
        return int(self.seq_length / self.block_size)

    def init(self, kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        desc = [
            f"bs={self.batch_size}",
            f"h={self.num_heads}",
            f"k={self.hidden_size}",
            f"seq={self.seq_length}",
            f"bl={self.block_size}",
        ]
        return ",".join(desc)


class AttentionMask(object):
    def __init__(self, config=None):
        super().__init__()
        if config is None:
            config = Configuration()
        self.config = config

    def is_blocked(self):
        return self.config.block_size != 1

    def is_valid_config(self, keep_blocked=True):
        return True

    def expand(self, mask):
        if mask.ndim == 2:
            return mask.unsqueeze(0).expand(self.config.num_heads, -1, -1)

    def gen_mask(self, keep_blocked=True):
        raise NotImplementedError("Abstract data class")

    def set_config_attr(self, key, value):
        setattr(self.config, key, value)

    def __str__(self):
        raise NotImplementedError("Abstract data type")

    def __call__(self):
        mask = self.gen_mask()
        return mask, self.config, str(self)


class RandomAttentionMask(AttentionMask):
    """
    This is a Random mask. Useful for performance and memory analysis.
    """

    def __init__(self, config=None):
        super(RandomAttentionMask, self).__init__(config)
        self.set_config_attr("mask_prob", 0.5)

    def gen_mask(self, keep_blocked=True):
        seq_length = self.config.seq_length
        if keep_blocked:
            seq_length = self.config.blocked_seq_length

        mask = torch.rand(seq_length, seq_length) > self.config.mask_prob
        return self.expand(mask)

    def __str__(self):
        return "random"


class LowerTriangularAttentionMask(AttentionMask):
    """
    This is a lower triangular mask. This is common in decoder only models.
    This should reduce the computation and memory to roughly half as close to
    half of the mask is zero.

    The mask stays same for each head and each input.

    Nit pick (TODO) - While blocking, we need to ensure that the blocks along
    the diagonals are themselves lower triangular blocks. But, for performance
    measurement, this is ok to ignore as we treat the whole block as useful
    values.
    """

    def __init__(self, config=None):
        super(LowerTriangularAttentionMask, self).__init__(config)

    def gen_mask(self, keep_blocked=True):
        seq_length = self.config.seq_length
        if keep_blocked:
            seq_length = self.config.blocked_seq_length
        return self.expand(causal_1d_pattern(seq_length))

    def __str__(self):
        return "lower_triangular"


class BigBirdAttentionMask(AttentionMask):
    """
    BigBird mask are composed of three types of masks - random, global and window.
    For more details, refer to https://arxiv.org/pdf/2007.14062.pdf

    One point to note is that mask is per head here. So, mask is 3D tensor.
    (num_heads, seq_length, seq_length).
    """

    def __init__(self, config=None):
        super(BigBirdAttentionMask, self).__init__(config)
        self.mask_per_head = True
        self.set_config_attr("num_global_tokens", 2 * self.config.block_size)
        self.set_config_attr("num_random_tokens", 3 * self.config.block_size)
        self.set_config_attr("num_window_tokens", 3 * self.config.block_size)

    def gen_global_mask(self, seq_length):
        # Global tokens are tokens that attend to all tokens and to whom all tokens attend to in the sequence
        num_global_blocks = self.config.num_global_tokens // self.config.block_size
        mask_indices = torch.randint(0, seq_length - 1, size=(num_global_blocks,))
        mask_indices = torch.unique(mask_indices)
        query_mask = torch.zeros(seq_length).to(dtype=torch.bool)
        query_mask.scatter_(0, mask_indices, True)
        return global_token_pattern(query_mask)

    def gen_random_mask(self, seq_length):
        # Each query token attends over r random number of tokens
        num_random_blocks = self.config.num_random_tokens // self.config.block_size
        mask_indices = torch.randint(
            0, seq_length - 1, size=(seq_length, num_random_blocks)
        )
        random_mask = torch.zeros(seq_length, seq_length).to(dtype=torch.bool)
        random_mask.scatter_(1, mask_indices, True)
        return random_mask

    def gen_window_mask(self, seq_length):
        num_window_blocks = self.config.num_window_tokens // self.config.block_size
        if num_window_blocks % 2 == 0:
            num_window_blocks += 1
        return local_1d_pattern(seq_length, num_window_blocks)

    def gen_mask(self, keep_blocked=True):
        seq_length = self.config.seq_length
        if keep_blocked:
            seq_length = self.config.blocked_seq_length
        assert keep_blocked, "Not implemented, call to_dense later to get full tensor"

        if self.mask_per_head:
            head_masks = []

            for _ in range(self.config.num_heads):
                global_mask = self.gen_global_mask(seq_length)
                random_mask = self.gen_random_mask(seq_length)
                window_mask = self.gen_window_mask(seq_length)

                mask = global_mask + random_mask + window_mask
                head_masks.append(mask)

            mask = torch.stack(head_masks)
        else:
            global_mask = self.gen_global_mask(seq_length)
            random_mask = self.gen_random_mask(seq_length)
            window_mask = self.gen_window_mask(seq_length)

            mask = global_mask + random_mask + window_mask
            mask = self.expand(mask)
        return mask

    def __str__(self):
        return "bigbird"


class AxialAttentionMask(AttentionMask):
    """
    BigBird mask are composed of three types of masks - random, global and window.
    For more details, refer to https://arxiv.org/pdf/2007.14062.pdf

    One point to note is that mask is per head here. So, mask is 3D tensor.
    (num_heads, seq_length, seq_length).
    """

    def __init__(self, config=None):
        super(AxialAttentionMask, self).__init__(config)
        if config is None:
            self.set_config_attr("seq_length", 1024)

    def is_valid_config(self, keep_blocked=True):
        seq_length = self.config.seq_length
        if keep_blocked:
            seq_length = self.config.blocked_seq_length
        H = int(math.sqrt(seq_length))
        if H * H == seq_length:
            return True
        return False

    def gen_mask(self, keep_blocked=True):
        seq_length = self.config.seq_length
        if keep_blocked:
            seq_length = self.config.blocked_seq_length
        H = int(math.sqrt(seq_length))
        assert H * H == seq_length, f"H={H}, seq_length={seq_length}"
        return self.expand(axial_2d_pattern(H, H))

    def __str__(self):
        return "axial"


class LocalAttentionMask(AttentionMask):
    """
    BigBird mask are composed of three types of masks - random, global and window.
    For more details, refer to https://arxiv.org/pdf/2007.14062.pdf

    One point to note is that mask is per head here. So, mask is 3D tensor.
    (num_heads, seq_length, seq_length).
    """

    def __init__(self, config=None):
        super(LocalAttentionMask, self).__init__(config)
        self.set_config_attr("num_local_blocks", 3)
        if config is None:
            self.set_config_attr("seq_length", 1024)

    def is_valid_config(self, keep_blocked=True):
        seq_length = self.config.seq_length
        if keep_blocked:
            seq_length = self.config.blocked_seq_length
        H = int(math.sqrt(seq_length))
        if H * H == seq_length:
            return True
        return False

    def gen_mask(self, keep_blocked=True):
        seq_length = self.config.seq_length
        if keep_blocked:
            seq_length = self.config.blocked_seq_length
        H = int(math.sqrt(seq_length))
        assert H * H == seq_length, f"H={H}, seq_length={seq_length}"
        return self.expand(local_2d_pattern(H, H, self.config.num_local_blocks))

    def __str__(self):
        return "local"


##############################################
# Class to organize the experiments
##############################################


class Experiment(object):
    def __init__(self, mode, dtype, do_accuracy_check, profile_sputnik):
        self.mode = mode
        self.dtype = dtype
        self.do_accuracy_check = do_accuracy_check
        self.profile_sputnik = profile_sputnik

    def reset_results(self):
        self.results = {}
        self.results["flops"] = {}
        self.results["time"] = {}
        self.results["memory"] = {}
        self.results["speedup"] = {}
        self.results["memory_savings"] = {}

    def do_mem(sel, fn):
        # bookeeping
        torch.cuda.empty_cache()
        torch.cuda.reset_peak_memory_stats()
        # actually run the function
        fn()
        fn()
        torch.cuda.synchronize()
        return torch.cuda.max_memory_allocated() // 2**20

    def gen_config(self):
        raise NotImplementedError("Not setup")

    def plot(self, sparsity, pattern_name):
        raise NotImplementedError("Not setup")

    def run(self):
        raise NotImplementedError("Not setup")

    def add_kv(self, d, d_key, d_value, testcase):
        d_value = max(0, d_value)
        if d_key not in d:
            d[d_key] = {}
        d[d_key][testcase.name] = d_value

    def bench_all(
        self, a, b, tests, mask_config, sparsity, baseline_name, op_flops, dict_key
    ):

        if self.do_accuracy_check:
            self.check_all(tests, a, b)

        for testcase in tests:
            gc.collect()
            torch.cuda.empty_cache()
            torch.cuda.reset_peak_memory_stats()
            try:
                fn = testcase.prepare_callable(a, b, testcase.mask, testcase.config)
                ms = triton.testing.do_bench(fn)[0]
                flops = op_flops / ms * 1e3  # TFlop per second
                mem = self.do_mem(fn)
            except Exception:
                # raise
                ms = -1
                flops = -1
                mem = -1

            # Write into results
            # dict_key = f"sp={sparsity}%,{mask_config}"
            self.add_kv(self.results["time"], dict_key, ms, testcase)
            self.add_kv(self.results["flops"], dict_key, flops, testcase)
            self.add_kv(self.results["memory"], dict_key, mem, testcase)

            speedup = self.results["time"][dict_key][baseline_name] / ms
            memory_savings = self.results["memory"][dict_key][baseline_name] / mem
            self.add_kv(self.results["speedup"], dict_key, speedup, testcase)
            self.add_kv(self.results["flops"], dict_key, flops, testcase)
            self.add_kv(
                self.results["memory_savings"], dict_key, memory_savings, testcase
            )

            desc = f"sparsity={sparsity}, ops={op_flops}, time={ms}, tflops={flops}, mem={mem}"
            print(f"{testcase.name} --> {mask_config}, {desc}")

    def get_inputs(self, config, device="cuda"):
        # if mode = sddmm, a, b = query, key
        # if mode = spmm, a, b = attn, value
        if self.mode == "sddmm":
            return [
                torch.randn(
                    config.batch_size,
                    config.num_heads,
                    config.seq_length,
                    config.hidden_size // config.num_heads,
                    device=device,
                    dtype=self.dtype,
                )
                for _ in range(2)
            ]
        else:
            assert self.mode == "spmm"
            attn = torch.randn(
                config.batch_size,
                config.num_heads,
                config.seq_length,
                config.seq_length,
                device=device,
                dtype=self.dtype,
            )
            value = torch.randn(
                config.batch_size,
                config.num_heads,
                config.seq_length,
                config.hidden_size // config.num_heads,
                device=device,
                dtype=self.dtype,
            )
            return [attn, value]

    def torch_matmul_callable(self, a, b, mask, config):
        input_a = mask_tensor(a, mask, config) if self.mode == "spmm" else a
        input_b = b.transpose(-1, -2) if self.mode == "sddmm" else b

        def torch_fn():
            return torch.matmul(input_a, input_b)

        return torch_fn

    def get_triton_fn(self, mask, config, mode="sddmm"):
        if mode == "sddmm":
            return blocksparse_matmul(
                layout=mask,
                block=config.block_size,
                mode="sdd",
                device="cuda",
                trans_a=False,
                trans_b=True,
            )
        else:
            assert mode == "spmm"
            return blocksparse_matmul(
                layout=mask,
                block=config.block_size,
                mode="dsd",
                device="cuda",
                trans_a=False,
                trans_b=False,
            )

    def triton_callable(self, a, b, mask, config):
        triton_kernel = self.get_triton_fn(mask, config, self.mode)
        input_a = sparsify_tensor(a, mask, config) if self.mode == "spmm" else a
        input_b = b

        def triton_fn():
            return triton_kernel(input_a, input_b)

        return triton_fn

    def prepare_sputnik_inputs(self, query, key, config, mask):
        # - sparse / sputnik
        mask_cs = torch.ones(
            [config.batch_size, config.num_heads, config.seq_length, config.seq_length],
            dtype=torch.bool,
            device="cuda",
        )
        mask_cs = triton.testing.mask_tensor(
            mask_cs, mask, config.block_size, value=False
        )
        # Sputnik kernels only handle fp32
        query_cs = query.flatten(start_dim=0, end_dim=1).to(torch.float32)
        key_cs = key.flatten(start_dim=0, end_dim=1).to(torch.float32)

        query_cs = query_cs.contiguous()
        key_cs = key_cs.transpose(-2, -1)

        sparse_mask_cs = SparseCS(
            mask_cs.flatten(start_dim=0, end_dim=1).contiguous(),
            device=torch.device("cuda"),
        )
        return query_cs, key_cs, sparse_mask_cs

    def sputnik_callable(self, a, b, mask, config):
        assert self.mode == "sddmm"
        a_cs, b_cs, sparse_mask_cs = self.prepare_sputnik_inputs(a, b, config, mask)

        def sputnik_fn():
            return _matmul_with_mask(a_cs, b_cs, sparse_mask_cs)

        return sputnik_fn

    def get_op_flops(self, mask, config):
        # Measure total compute ops
        op_flops = (
            2  # FMA
            * config.batch_size  # batched matmul
            * (config.hidden_size // config.num_heads)  # Reduce dimension
            * float(mask.sum())
            * config.block_size
            * config.block_size  # Effective seq length * seq_length
            * 1e-12  # TFlops
        )
        return op_flops

    def check_all(self, tests, a, b):
        ref_test = tests[0]
        ref_out = ref_test.prepare_callable(a, b, ref_test.mask, ref_test.config)()

        res_test = tests[1]
        res_out = res_test.prepare_callable(a, b, res_test.mask, res_test.config)()

        self.check_accuracy(ref_out, res_out, ref_test.mask, ref_test.config)

    def check_accuracy(self, ref_full, res_bsr, mask, config):
        if self.mode == "sddmm":
            # Get the dense representation of the bsr tensor
            # Use triton sparse * dense multiplication to get the dense tensor back
            sparse_dot_dsd = blocksparse_matmul(
                layout=mask,
                block=config.block_size,
                mode="dsd",
                device="cuda",
                trans_a=False,
                trans_b=False,
            )
            identity = torch.eye(
                config.seq_length, config.seq_length, device=device, dtype=self.dtype
            )

            identity = identity.expand(config.batch_size, config.num_heads, -1, -1)
            res = sparse_dot_dsd(res_bsr, identity)

            # Get the res where values are masked. Expand the blocked mask
            # ref = triton.testing.mask_tensor(ref_full, mask, config.block_size)
            full_mask = densify_mask(mask, config)
            ref = ref_full * full_mask.to(dtype=self.dtype, device=device)
            try:
                assert torch.allclose(ref, res, atol=1e-3, rtol=1e-3)
            except RuntimeError:
                pass
            except AssertionError:
                raise
        else:
            assert self.mode == "spmm"
            # Both are dense outputs
            try:
                assert torch.allclose(ref_full, res_bsr, atol=1e-3, rtol=1e-3)
            except RuntimeError:
                pass
            except AssertionError:
                import pdb

                pdb.set_trace()
                raise


class DifferentPatternExperiment(Experiment):
    """
    In this experiment, we check if sparsity pattern (like bigbird, lower triangular
    etc) changes the performance of different kernels. The idea is to check if
    changing sparsity pattern, while keeping total sparsity ratio same, leads to any
    perforamnce differences.

    We will perform two experiments

    1) LowerTraingularMask vs RandomMask - Both have ~50% sparsity.
    2) BigBird Mask vs RandomMask - Both have same sparsity.
    """

    def __init__(self, mode, dtype, do_accuracy_check, profile_sputnik=False):
        super(DifferentPatternExperiment, self).__init__(
            mode, dtype, do_accuracy_check, profile_sputnik
        )

    def gen_config(self):
        batch_sizes = [32]
        heads = [16]
        seq_lengths = [1024, 2048]
        block_sizes = [64]
        hidden_sizes = [1024, 4096, 8192]
        for batch in batch_sizes:
            for hidden_size in hidden_sizes:
                for head in heads:
                    for seq in seq_lengths:
                        for block in block_sizes:
                            entry = {
                                "batch_size": batch,
                                "num_heads": head,
                                "seq_length": seq,
                                "block_size": block,
                                "hidden_size": hidden_size,
                            }
                            yield entry

    def plot(self, sparsity, config, pattern_name):
        desc = [
            f"bs={config.batch_size}",
            f"nheads={config.num_heads}",
            f"block={config.block_size}",
            f"dtype={self.dtype}",
        ]
        title_suffix = ",".join(desc)
        pretty_barplot(
            self.results["speedup"],
            title=f"{self.mode} - Pattern experiment ({sparsity}%) - speedup\n"
            + title_suffix,
            filename=f"same_sparsity_{self.mode}_{self.dtype}_{pattern_name}_time.svg",
            dash_key="pytorch",
            units="Speedup normalized to torch_matmul",
        )

        pretty_barplot(
            self.results["flops"],
            title=f"{self.mode} - Pattern experiment ({sparsity}%) - throughput\n"
            + title_suffix,
            filename=f"same_sparsity_{self.mode}_{self.dtype}_{pattern_name}_flops.svg",
            dash_key="pytorch",
            units="TFlops/s",
        )

        pretty_barplot(
            self.results["memory_savings"],
            title=f"{self.mode} - Pattern experiment ({sparsity}%) - memory savings\n"
            + title_suffix,
            filename=f"same_sparsity_{self.mode}_{self.dtype}_{pattern_name}_memory.svg",
            dash_key="pytorch",
            units="Memory savings normalized to torch_matmul",
        )

    def run(self):
        for MaskGenType in [LowerTriangularAttentionMask, BigBirdAttentionMask]:
            self.reset_results()
            for config in self.gen_config():
                # Get pattern mask
                pattern_mask, pattern_config, pattern_name = get_mask(
                    MaskGenType, config
                )
                sparsity = get_sparsity(pattern_mask)
                mask_prob = sparsity / 100

                # Get random mask
                random_mask, random_config, _ = get_mask(
                    RandomAttentionMask,
                    config,
                    [("mask_prob", mask_prob)],
                )
                print(f"{pattern_name} sparsity", get_sparsity(pattern_mask))
                print("Random sparsity", get_sparsity(random_mask))

                # Create input tensors
                a, b = self.get_inputs(random_config)

                tests = []
                baseline_name = "torch-matmul"
                tests.append(
                    TestCase(
                        self.torch_matmul_callable,
                        random_mask,
                        random_config,
                        f"{baseline_name}",
                    )
                )
                tests.append(
                    TestCase(
                        self.triton_callable,
                        random_mask,
                        random_config,
                        "triton-random",
                    )
                )
                tests.append(
                    TestCase(
                        self.triton_callable,
                        pattern_mask,
                        pattern_config,
                        f"triton-{pattern_name}",
                    )
                )
                if self.profile_sputnik and self.mode == "sddmm":
                    tests.append(
                        TestCase(
                            self.sputnik_callable,
                            random_mask,
                            random_config,
                            "sputnik-random",
                        )
                    )
                    tests.append(
                        TestCase(
                            self.sputnik_callable,
                            pattern_mask,
                            pattern_config,
                            f"sputnik-{pattern_name}",
                        )
                    )

                dict_key = f"hidden={random_config.hidden_size},seq_len={random_config.seq_length}"
                self.bench_all(
                    a,
                    b,
                    tests,
                    random_config,
                    sparsity,
                    baseline_name,
                    self.get_op_flops(random_mask, random_config),
                    dict_key,
                )

                ideal_testcase = TestCase(None, None, None, "Ideal")
                ideal_speedup = round(100 / (100 - sparsity), 1)
                self.add_kv(
                    self.results["speedup"], dict_key, ideal_speedup, ideal_testcase
                )
                self.add_kv(
                    self.results["memory_savings"],
                    dict_key,
                    ideal_speedup,
                    ideal_testcase,
                )
                self.plot(sparsity, random_config, pattern_name)


class VarySparsityExperiment(Experiment):
    """
    In this experiment, we check how sparsity ration affects the performance.
    """

    def __init__(self, mode, dtype, do_accuracy_check, profile_sputnik=False):
        super(VarySparsityExperiment, self).__init__(
            mode, dtype, do_accuracy_check, profile_sputnik
        )

    def gen_config(self):
        batch_sizes = [32]
        heads = [16]
        seq_lengths = [2048]
        hidden_sizes = [1024, 8192]
        block_sizes = [64]
        for batch in batch_sizes:
            for seq in seq_lengths:
                for head in heads:
                    for block in block_sizes:
                        for hidden_size in hidden_sizes:
                            entry = {
                                "batch_size": batch,
                                "num_heads": head,
                                "seq_length": seq,
                                "block_size": block,
                                "hidden_size": hidden_size,
                            }
                            yield entry

    def plot(self, sparsity, config, pattern_name):
        desc = [
            f"bs={config.batch_size}",
            f"nheads={config.num_heads}",
            f"block={config.block_size}",
            f"dtype={self.dtype}",
            f"seq_len={config.seq_length}",
        ]
        title_suffix = ",".join(desc)
        pretty_barplot(
            self.results["speedup"],
            title=f"{self.mode} - SparsityRatio experiment speedup\n" + title_suffix,
            filename=f"vary_sparsity_{self.mode}_{self.dtype}_{pattern_name}_time.svg",
            dash_key="pytorch",
            units="Speedup normalized to torch_matmul",
        )

        pretty_barplot(
            self.results["flops"],
            title=f"{self.mode} - SparsityRatio experiment throughput\n" + title_suffix,
            filename=f"vary_sparsity_{self.mode}_{self.dtype}_{pattern_name}_flops.svg",
            dash_key="pytorch",
            units="TFlops/s",
        )

        pretty_barplot(
            self.results["memory_savings"],
            title=f"{self.mode} - SparsityRatio experiment memory savings\n"
            + title_suffix,
            filename=f"vary_sparsity_{self.mode}_{self.dtype}_{pattern_name}_memory.svg",
            dash_key="pytorch",
            units="Memory savings normalized to torch_matmul",
        )

    def run(self):
        self.reset_results()
        random_config = None
        for config in self.gen_config():
            for x in range(10, 100, 10):
                mask_prob = x / 100.0
                # Get random mask
                random_mask, random_config, _ = get_mask(
                    RandomAttentionMask,
                    config,
                    [
                        ("mask_prob", mask_prob),
                    ],
                )
                sparsity = get_sparsity(random_mask)
                print("Random sparsity", get_sparsity(random_mask))

                # Create input tensors
                a, b = self.get_inputs(random_config)

                tests = []
                baseline_name = "torch-matmul"
                tests.append(
                    TestCase(
                        self.torch_matmul_callable,
                        random_mask,
                        random_config,
                        f"{baseline_name}",
                    )
                )
                tests.append(
                    TestCase(
                        self.triton_callable,
                        random_mask,
                        random_config,
                        "triton-random",
                    )
                )
                if self.profile_sputnik and self.mode == "sddmm":
                    tests.append(
                        TestCase(
                            self.sputnik_callable,
                            random_mask,
                            random_config,
                            "sputnik-random",
                        )
                    )
                dict_key = f"sp={mask_prob},hidden={random_config.hidden_size}"
                self.bench_all(
                    a,
                    b,
                    tests,
                    random_config,
                    sparsity,
                    baseline_name,
                    self.get_op_flops(random_mask, random_config),
                    dict_key,
                )

                ideal_testcase = TestCase(None, None, None, "Ideal")
                ideal_speedup = round(100 / (100 - mask_prob * 100), 1)
                self.add_kv(
                    self.results["speedup"], dict_key, ideal_speedup, ideal_testcase
                )
                self.add_kv(
                    self.results["memory_savings"],
                    dict_key,
                    ideal_speedup,
                    ideal_testcase,
                )
        self.plot(None, random_config, "random")


class BlockSizeExperiment(Experiment):
    """
    In this experiment, we analyze how increasing the block size affects
    performance.  We will take the lower triangular pattern. As we increase the
    batch size, the blocks near the diagonal have to do more unnecessary
    computation (the effective sparsity starts decreasing).
    """

    def __init__(self, mode, dtype, do_accuracy_check, profile_sputnik=False):
        super(BlockSizeExperiment, self).__init__(
            mode, dtype, do_accuracy_check, profile_sputnik
        )

    def gen_config(self):
        batch_sizes = [32]
        heads = [16]
        seq_lengths = [2048]
        block_sizes = [32, 64, 128, 256]
        hidden_sizes = [1024, 8192]
        for batch in batch_sizes:
            for seq in seq_lengths:
                for hidden_size in hidden_sizes:
                    for block in block_sizes:
                        for head in heads:
                            entry = {
                                "batch_size": batch,
                                "num_heads": head,
                                "seq_length": seq,
                                "block_size": block,
                                "hidden_size": hidden_size,
                            }
                            yield entry

    def plot(self, sparsity, config, pattern_name):
        pretty_barplot(
            self.results["speedup"],
            title=f"{self.mode} - BlockSize experiment speedup\n"
            f"bs={config.batch_size}, nheads={config.num_heads}, seq_len={config.seq_length}, dtype={self.dtype}",
            filename=f"vary_block_size_{self.mode}_{self.dtype}_{pattern_name}_time.svg",
            dash_key="pytorch",
            units="Speedup normalized to torch matmul",
        )

        pretty_barplot(
            self.results["flops"],
            title=f"{self.mode} - BlockSize experiment throughput\n"
            f"bs={config.batch_size}, nheads={config.num_heads}, seq_len={config.seq_length}, dtype={self.dtype}",
            filename=f"vary_block_size_{self.mode}_{self.dtype}_{pattern_name}_flops.svg",
            dash_key="pytorch",
            units="TFlops/s",
        )

        pretty_barplot(
            self.results["memory_savings"],
            title=f"{self.mode} - BlockSize experiment memory savings\n"
            f"bs={config.batch_size}, nheads={config.num_heads}, seq_len={config.seq_length}, dtype={self.dtype}",
            filename=f"vary_block_size_{self.mode}_{self.dtype}_{pattern_name}_memory.svg",
            dash_key="pytorch",
            units="Memory savings normalized to torch matmul",
        )

    def get_op_flops(self, mask, config):
        # Op flops here refer to the original non blocked attention mask, where
        # no unnecessary elements are unmasked. We can compute this by computing
        # the total flops of batch matmul and then multiply by (n+1)/2n.
        num_masked_elems = (config.seq_length + 1) / (2.0 * config.seq_length)
        op_flops = (
            2  # FMA
            * config.batch_size  # batched matmul
            * config.num_heads
            * (config.hidden_size // config.num_heads)  # Reduce dimension
            * config.seq_length
            * config.seq_length
            * num_masked_elems
            * 1e-12  # TFlops
        )
        return op_flops

    def run(self):
        self.reset_results()
        lt_config = None
        for config in self.gen_config():
            lt_mask, lt_config, lt_name = get_mask(
                LowerTriangularAttentionMask,
                config,
            )
            sparsity = get_sparsity(lt_mask)
            print("Effective sparsity", sparsity)
            if lt_config.seq_length == 2048:
                plot_mask(lt_mask, lt_config, f"lt_mask_{lt_config.block_size}.svg")

            # Create input tensors
            a, b = self.get_inputs(lt_config)

            tests = []
            baseline_name = "torch-matmul"
            tests.append(
                TestCase(
                    self.torch_matmul_callable, lt_mask, lt_config, f"{baseline_name}"
                )
            )
            tests.append(
                TestCase(self.triton_callable, lt_mask, lt_config, "triton-random")
            )
            if self.profile_sputnik and self.mode == "sddmm":
                tests.append(
                    TestCase(
                        self.sputnik_callable, lt_mask, lt_config, "sputnik-random"
                    )
                )
            dict_key = f"hidden={lt_config.hidden_size}, block={lt_config.block_size}"
            self.bench_all(
                a,
                b,
                tests,
                lt_config,
                sparsity,
                baseline_name,
                self.get_op_flops(lt_mask, lt_config),
                dict_key,
            )

            ideal_testcase = TestCase(None, None, None, "Ideal")
            seq_len = lt_config.seq_length
            total_elems = seq_len * seq_len
            nnz = seq_len * (seq_len + 1) / 2
            ideal_speedup = (1.0 * total_elems) / nnz

            self.add_kv(
                self.results["speedup"], dict_key, ideal_speedup, ideal_testcase
            )
            self.add_kv(
                self.results["memory_savings"],
                dict_key,
                ideal_speedup,
                ideal_testcase,
            )
        self.plot(None, lt_config, lt_name)


if __name__ == "__main__":
    for MaskGen in [
        RandomAttentionMask,
        LowerTriangularAttentionMask,
        BigBirdAttentionMask,
        AxialAttentionMask,
        LocalAttentionMask,
    ]:
        mask_gen = MaskGen()
        mask, config, name = mask_gen()
        plot_mask(mask, config, f"{name}.svg")

    for mode in ["sddmm", "spmm"]:
        DifferentPatternExperiment(mode, torch.float16, True).run()
        VarySparsityExperiment(mode, torch.float16, True).run()
        BlockSizeExperiment(mode, torch.float16, True).run()



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from pathlib import Path
from typing import Any, Callable, Dict, Set, Union

from xformers.utils import (
    generate_matching_config,
    get_registry_decorator,
    import_all_modules,
)

from .base import PositionEmbedding, PositionEmbeddingConfig  # noqa

# CREDITS: Classy Vision registry mechanism

POSITION_EMBEDDING_REGISTRY: Dict[str, Any] = {}
POSITION_EMBEDDING_CLASS_NAMES: Set[str] = set()


def build_positional_embedding(config: Union[Dict[str, Any], PositionEmbeddingConfig]):
    """Builds a position encoding from a config.

    This assumes a 'name' key in the config which is used to determine what
    attention class to instantiate. For instance, a config `{"name": "my_position_encoding",
    "foo": "bar"}` will find a class that was registered as "my_position_encoding"
    (see :func:`register_positional_embedding`) and call .from_config on it."""

    if not isinstance(config, PositionEmbeddingConfig):
        config_instance = generate_matching_config(
            config, POSITION_EMBEDDING_REGISTRY[config["name"]].config
        )
    else:
        config_instance = config

    return POSITION_EMBEDDING_REGISTRY[config_instance.name].constructor.from_config(
        config_instance
    )


"""Registers a PositionEncoding subclass.

    This decorator allows xFormers to instantiate a subclass of PositionEncoding
    from a configuration file, even if the class itself is not part of the
    xFormers framework. To use it, apply this decorator to a `PositionEncoding`
    subclass, like this:

    .. code-block:: python

        @dataclass
        class MyConfig:
            ...

        @register_positional_embedding('my_encoding', MyConfig)
        class MyEncoding(PositionEncoding):
            ...

    To instantiate a position encoding from a configuration file, see :func:`build_positional_embedding`."""
register_positional_embedding: Callable[
    [str, Any], Callable[[Any], Any]
] = get_registry_decorator(
    POSITION_EMBEDDING_REGISTRY,
    POSITION_EMBEDDING_CLASS_NAMES,
    PositionEmbedding,
    PositionEmbeddingConfig,
)


from .rotary import RotaryEmbedding  # noqa
from .sine import SinePositionalEmbedding  # type: ignore  # noqa
from .vocab import VocabEmbedding  # noqa

__all__ = [
    "RotaryEmbedding",
    "SinePositionalEmbedding",
    "VocabEmbedding",
    "build_positional_embedding",
    "register_positional_embedding",
]

# automatically import any Python files in the directory
import_all_modules(
    str(Path(__file__).parent), "xformers.components.positional_embedding"
)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import torch


def _coo_to_csr(m, n, row_indices, column_indices):
    # assumes coalesced coo
    row_offsets = row_indices.bincount(minlength=n).cumsum(0, dtype=row_indices.dtype)
    row_offsets = torch.nn.functional.pad(row_offsets, (1, 0))
    return row_offsets, column_indices


def _csr_to_coo(m, n, row_offsets, column_indices):
    # convert from compressed rows to uncompressed
    indices = torch.arange(m, dtype=row_offsets.dtype, device=row_offsets.device)
    row_sizes = torch.diff(row_offsets)
    row_coo = torch.repeat_interleave(indices, row_sizes.long())
    return row_coo, column_indices


def _diffsort(a):
    return torch.argsort(torch.diff(a), dim=0, descending=True)


def _get_transpose_info(m, n, row_indices, row_offsets, column_indices):
    # strategy:
    # - uncompress the rows to have data in COO format
    # - get permutation for stable sort of the columns to get the rows for the transposed matrix
    # - compress the new rows and return the permutation to be applied on the values

    # convert from compressed rows to uncompressed
    row_coo, _ = _csr_to_coo(m, n, row_offsets, column_indices)

    # get the permutation for the stable sort
    row_offsets_t, perm = column_indices.sort(dim=0, stable=True)
    column_indices_t = row_coo[perm]

    row_offsets_t, _ = _coo_to_csr(m, n, row_offsets_t, column_indices)
    row_indices_t = _diffsort(row_offsets_t).int()

    return row_indices_t, row_offsets_t, column_indices_t, perm


def _transpose_with_info(values, _transpose_info):
    row_indices_t, row_offsets_t, column_indices_t, perm = _transpose_info
    values_t = values[:, perm]
    return row_indices_t, values_t, row_offsets_t, column_indices_t


def _transpose(m, n, row_indices, values, row_offsets, column_indices):
    _transpose_info = _get_transpose_info(
        m, n, row_indices, row_offsets, column_indices
    )
    return _transpose_with_info(values, _transpose_info)


def _nonzero_mask_to_sparse_csr_indices(mask, device):
    """Converts dense 2d matrix to a csr sparse matrix."""

    assert len(mask.shape) == 2
    index_dtype = torch.int32

    # Calculate the offset of each row.
    row_offsets = mask.sum(dim=-1, dtype=index_dtype).cumsum(dim=-1, dtype=index_dtype)
    row_offsets = torch.nn.functional.pad(row_offsets, (1, 0))

    # Create the row indices and sort them.
    row_indices = _diffsort(row_offsets).to(index_dtype)

    # Extract the column indices for the nonzero values.
    column_indices = torch.where(mask)[1].to(index_dtype).contiguous()

    row_indices = row_indices.to(device)
    row_offsets = row_offsets.to(device)
    column_indices = column_indices.to(device)
    return row_indices, row_offsets, column_indices


def _dense_to_sparse(matrix, device):
    """Converts dense 2d matrix to a csr sparse matrix."""

    assert len(matrix.shape) == 2
    value_dtype = torch.float32

    # Extract the nonzero values.
    mask = matrix != 0
    values = matrix[mask].to(dtype=value_dtype, device=device)

    row_indices, row_offsets, column_indices = _nonzero_mask_to_sparse_csr_indices(
        mask, device
    )
    return values, row_indices, row_offsets, column_indices


def _round_nnz(mask, divisible_by=4):
    nonzero = torch.where(mask)
    nnz = nonzero[0].shape[0]
    nonzero = tuple(n[: (nnz - nnz % divisible_by)] for n in nonzero)
    nm = torch.zeros_like(mask)
    nm[nonzero] = True
    return nm


def _dense3d_to_sparse(matrix, device):
    assert len(matrix.shape) == 3
    mask = matrix != 0
    if not torch.all(mask == mask[0]):
        raise ValueError("Expected the same sparsity pattern over the batch dimension")

    # for now, our kernels assume that we have the number of
    # nnz to be divisible by 4
    mask = _round_nnz(mask[0], divisible_by=4)
    mask = mask[None].expand(matrix.shape)

    values = matrix[mask].reshape(matrix.shape[0], -1).to(device)
    row_indices, row_offsets, column_indices = _nonzero_mask_to_sparse_csr_indices(
        mask[0], device
    )
    return values, row_indices, row_offsets, column_indices



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import logging
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Optional, Union

import torch

from xformers.components import Activation
from xformers.components.feedforward import (
    Feedforward,
    FeedforwardConfig,
    register_feedforward,
)

_is_fairscale_available = True

try:
    import torch.distributed as dist
    from fairscale.nn import MOELayer, Top2Gate  # type: ignore

    from xformers.components.feedforward import MLP

except ImportError:
    logging.warning(
        "Either FairScale or torch distributed is not available, MixtureOfExperts will not be exposed."
        " Please install them if you would like to use MoE"
    )
    _is_fairscale_available = False


if _is_fairscale_available:

    # Credits: initially implemented in FairScale for sanity checking
    class RoundRobinGate(torch.nn.Module):
        def __init__(self, model_dim, num_experts):
            super().__init__()
            self.model_dim = model_dim
            self.num_experts = num_experts

        def forward(self, input):
            s = input.shape[0]
            assert s % self.num_experts == 0, f"{s} % {self.num_experts} != 0"
            capacity = 2 * s // self.num_experts
            output = torch.zeros(
                s, self.num_experts, capacity, dtype=input.dtype, device=input.device
            )
            for i in range(s):
                output[i, i % self.num_experts, i // self.num_experts] = 1.0
            return 0.0, output, output.bool()

    class GateConfig(str, Enum):
        RoundRobin = "round_robin"
        Top2 = "top_2"
        # Other gating techniques could be exposed here

    @dataclass
    class MoEConfig(FeedforwardConfig):
        number_of_experts: int
        gate: GateConfig
        number_of_local_experts: Optional[int] = None
        expert_constructor: Optional[Any] = None
        hidden_layer_multiplier: Optional[int] = None
        group: Optional[Any] = None

    @register_feedforward("MixtureOfExperts", MoEConfig)
    class MixtureOfExperts(Feedforward):
        """
        A MLP variant which uses the "Mixture of Experts" paradigm, as described in Gshard_.
        xFormers uses the FairScale_ implementation under the hood.

        .. warning: Please note that most of the benefits of MoE are present in a distributed training environmentt

        .. _Gshard: https://arxiv.org/pdf/2006.16668.pdf
        .. _FairScale: https://github.com/facebookresearch/fairscale/
        """

        def __init__(
            self,
            dim_model: int,
            dropout: float,
            activation: Activation,
            number_of_experts: int,
            gate: Union[GateConfig, torch.nn.Module],
            number_of_local_experts: Optional[int] = None,
            expert_constructor: Optional[Callable[[], torch.nn.Module]] = None,
            hidden_layer_multiplier: Optional[int] = None,
            group: Optional[Any] = None,
            *_,
            **__,
        ):
            super().__init__()

            # Handle a possibly uninitialized process group
            assert (
                dist.is_initialized()
            ), "Mixture of Experts require torch distributed to be initialized"

            if number_of_local_experts is not None:
                assert number_of_experts >= number_of_local_experts
            else:
                if dist.get_world_size() == 1:
                    logging.warning("Local experts no specified but world size of 1")
                    logging.warning("Assuming that all experts are local")
                    number_of_local_experts = number_of_experts
                else:
                    number_of_local_experts = 1

            # Programatically handle the gating technique
            if not isinstance(gate, torch.nn.Module):
                gate_constructor = {
                    GateConfig.RoundRobin: RoundRobinGate,
                    GateConfig.Top2: Top2Gate,
                }[gate]

                self.gate = gate_constructor(dim_model, number_of_experts)
            else:
                self.gate = gate

            # Programatically handle the experts
            if expert_constructor is None:

                multiplier = (
                    hidden_layer_multiplier
                    if hidden_layer_multiplier is not None
                    else 4
                )

                def expert_constructor() -> torch.nn.Module:
                    return MLP(dim_model, dropout, activation, multiplier)

                assert expert_constructor is not None

            local_experts = torch.nn.ModuleList(
                [expert_constructor() for _ in range(number_of_local_experts)]
            )

            self.moe = MOELayer(gate=self.gate, experts=local_experts, group=group)

            self.requires_cuda = True

        def forward(self, inputs: torch.Tensor) -> torch.Tensor:
            # FairScale MoE assumes that the dimensions are [S, B, E]
            # xFormers assumes [B, S, E]
            return self.moe(inputs.movedim(0, 1)).movedim(0, 1)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from typing import Optional

import torch


# Reshapes key padding mask from (batch_size, src_len) -> (batch_size * num_heads 1, src_len)
def reshape_key_padding_mask(
    key_padding_mask: torch.Tensor, batched_dim: int
) -> torch.Tensor:
    assert key_padding_mask.ndim == 2
    batch_size, src_len = key_padding_mask.size()
    num_heads = batched_dim // batch_size
    return _reshape_key_padding_mask(key_padding_mask, batch_size, src_len, num_heads)


def _reshape_key_padding_mask(
    key_padding_mask: torch.Tensor, batch_size: int, src_len: int, num_heads: int
) -> torch.Tensor:
    assert key_padding_mask.shape == (batch_size, src_len)
    key_padding_mask = (
        key_padding_mask.view(batch_size, 1, 1, src_len)
        .expand(-1, num_heads, -1, -1)
        .reshape(batch_size * num_heads, 1, src_len)
    )
    return key_padding_mask


# Combine the attention mask and key padding mask into a single mask
# Taken from https://github.com/pytorch/pytorch/blob/master/torch/nn/functional.py
# Additive masking not yet supported
def maybe_merge_masks(
    att_mask: Optional[torch.Tensor],
    key_padding_mask: Optional[torch.Tensor],
    batch_size: int,
    src_len: int,
    num_heads: int,
    tgt_len: Optional[int] = None,
) -> Optional[torch.Tensor]:
    if tgt_len is None:
        tgt_len = src_len
    if key_padding_mask is not None:
        assert key_padding_mask.shape == (batch_size, src_len)
        key_padding_mask = _reshape_key_padding_mask(
            key_padding_mask, batch_size, src_len, num_heads
        )
        if att_mask is None:
            # make sure dimensions of key padding mask are the same as those expected for att_mask
            att_mask = key_padding_mask.expand(-1, tgt_len, -1)
        # Assumption is that False means to mask.
        elif att_mask.dtype == torch.bool:
            att_mask = att_mask.logical_and(key_padding_mask)
        else:
            att_mask = att_mask.masked_fill(~key_padding_mask, float("-inf"))

    return att_mask


# Assumes that matrix passed in has had softmax applied to it.
def iterative_pinv(softmax_mat: torch.Tensor, n_iter=6, pinverse_original_init=False):
    """
    Computing the Moore-Penrose inverse.
    Use an iterative method from (Razavi et al. 2014) to approximate the Moore-Penrose inverse via efficient
    matrix-matrix multiplications.
    """

    i = torch.eye(
        softmax_mat.size(-1), device=softmax_mat.device, dtype=softmax_mat.dtype
    )
    k = softmax_mat

    # The entries of K are positive and ||K||_{\infty} = 1 due to softmax
    if pinverse_original_init:
        # This original implementation is more conservative to compute coefficient of Z_0.
        v = 1 / torch.max(torch.sum(k, dim=-2)) * k.transpose(-1, -2)
    else:
        # This is the exact coefficient computation, 1 / ||K||_1, of initialization of Z_0, leading to faster
        # convergence.
        v = (
            1
            / torch.max(torch.sum(k, dim=-2), dim=-1).values[:, None, None]
            * k.transpose(-1, -2)
        )

    for _ in range(n_iter):
        kv = torch.matmul(k, v)
        v = torch.matmul(
            0.25 * v,
            13 * i - torch.matmul(kv, 15 * i - torch.matmul(kv, 7 * i - kv)),
        )
    return v


def bool_mask_to_additive(
    mask: torch.Tensor, dtype: Optional[torch.dtype] = torch.float32
) -> torch.Tensor:
    assert (
        mask.dtype == torch.bool
    ), "This util is meant to convert in between bool masks and additive ones"

    mask_ = torch.zeros_like(mask, dtype=dtype)
    mask_[~mask] = float("-inf")
    return mask_



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import torch
from torch.cuda.amp import autocast

from xformers.components.attention import Attention, AttentionConfig, register_attention


@register_attention("fourier_mix", AttentionConfig)
class FourierMix(Attention):
    def __init__(self, dropout: float, *_, **__):
        """
        FFT-based pseudo-attention mechanism, from
        "
        "FNet: Mixing Tokens with Fourier Transforms"
        Lee-Thorp et al., 2021, https://arxiv.org/pdf/2105.03824.pdf
        """
        super().__init__()
        self.attn_drop = torch.nn.Dropout(dropout, inplace=False)

        # Properties specific to this attention mechanism
        self.supports_attention_mask = False
        self.requires_input_projection = False

    def forward(self, q: torch.Tensor, *_, **__):
        # Guard against autocast / fp16, not supported by torch.fft.fft2
        with autocast(enabled=False):
            att = torch.fft.fft2(q).real

        att = self.attn_drop(att)

        return att



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import json
import time
from contextlib import suppress
from typing import Any, Dict, List, Optional

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import torch
import torch.nn.functional as F
from sklearn.model_selection import ParameterGrid

# CREDITS: Sean Naren
from torch.autograd.profiler import record_function
from tqdm import tqdm

from xformers.components import Activation
from xformers.components.attention import ATTENTION_REGISTRY
from xformers.factory.block_factory import xFormerEncoderBlock, xFormerEncoderConfig

_use_cuda = torch.cuda.is_available()
_GLOBAL_ATTENTION_RATIO = 0.1  # arbitrary


def _get_attention_query_mask(sequence_length: int, ratio: float):
    mask = torch.rand((sequence_length, 1)) < ratio
    while torch.count_nonzero(mask) / float(mask.numel()) > ratio:
        mask = torch.rand((sequence_length, 1)) < ratio

    return mask


def _get_trace_handler(name: str):
    def trace_handler(prof):
        prof.export_chrome_trace(f"profile_{name}.json")
        prof.export_stacks(f"stacks_{name}.txt", "self_cuda_time_total")

    return trace_handler


def _train_for_several_steps(
    block: xFormerEncoderBlock,
    num_steps: int,
    batch_size: int,
    sequence_length: int,
    embed_dim: int,
    autocast: bool,
    device: torch.device,
    lr: float = 0.01,
    norm_type: Optional[float] = None,
    profile: bool = False,
    att_name: str = "",
) -> Dict[str, float]:
    # use SGD with momentum instead of Adam, since Adam is scale invariant
    # and this makes it bad for tests
    optim = torch.optim.SGD(block.parameters(), lr=lr, momentum=0.9)

    if _use_cuda:
        torch.cuda.empty_cache()
        torch.cuda.reset_peak_memory_stats()
        torch.cuda.synchronize()

    start_time = time.time()

    # Optional profiler, requires a context and some setup
    profiler = (
        torch.profiler.profile(  # type: ignore
            activities=[
                torch.profiler.ProfilerActivity.CPU,  # type: ignore
                torch.profiler.ProfilerActivity.CUDA,  # type: ignore
            ],
            schedule=torch.profiler.schedule(wait=1, warmup=1, active=1),  # type: ignore
            on_trace_ready=_get_trace_handler(
                f"{att_name}_batch_{batch_size}_seq_{sequence_length}_embed_dim_{embed_dim}"
            ),
            profile_memory=True,
            with_stack=True,
        )
        if profile
        else suppress()
    )

    # Actual vanilla training loop
    # - nonsensical data, but remove that from the compute time
    inputs = torch.rand(batch_size, sequence_length).to(device)

    with profiler as p:  # type: ignore
        for _ in range(num_steps):
            optim.zero_grad()

            with torch.cuda.amp.autocast(enabled=autocast):
                with record_function("attention_forward"):
                    output = block(inputs)

                with record_function("loss"):
                    loss = F.mse_loss(
                        inputs.unsqueeze(-1).repeat(1, 1, output.shape[-1]),
                        output,
                        reduction="sum",
                    )

            with record_function("backward"):
                loss.backward()

            if norm_type is not None:
                clip_norm = 0.3
                torch.nn.utils.clip_grad_norm_(block.parameters(), clip_norm, norm_type)
            optim.step()

            if p:
                p.step()

    if _use_cuda:
        torch.cuda.synchronize()
        max_memory = torch.cuda.max_memory_allocated() / 2**20
    else:
        max_memory = -1
    run_time = time.time() - start_time

    return {"run_time": run_time, "max_memory": round(max_memory, 1)}


def benchmark_model(num_warmup: int, num_steps: int, **kwargs) -> Dict[str, float]:
    # Run warm-up first
    warm_up_args = {**kwargs}
    warm_up_args["profile"] = False
    _train_for_several_steps(num_steps=num_warmup, **warm_up_args)

    return _train_for_several_steps(num_steps=num_steps, **kwargs)


def test_xformer_encoder_block(
    attention_name: str,
    feedforward_name: str,
    heads: int,
    attn_dropout: float,
    residual_dropout: float,
    causal: bool,
    activation: Activation,
    autocast: bool,
    batch_size: int,
    sequence_length: int,
    embed_dim: int,
    dropout: float,
    num_steps: int,
    num_warmup: int,
    device: torch.device,
    profile: bool,
) -> Dict[str, float]:

    block = instantiate_xformer(
        activation=activation,
        attention_name=attention_name,
        attn_dropout=attn_dropout,
        causal=causal,
        feedforward_name=feedforward_name,
        heads=heads,
        residual_dropout=residual_dropout,
        sequence_length=sequence_length,
        embed_dim=embed_dim,
        dropout=dropout,
    ).to(device)

    print(
        "Testing:",
        block,
        batch_size,
        sequence_length,
        embed_dim,
        autocast,
        device,
        attention_name,
    )

    return benchmark_model(
        num_steps=num_steps,
        num_warmup=num_warmup,
        block=block,
        batch_size=batch_size,
        sequence_length=sequence_length,
        embed_dim=embed_dim,
        autocast=autocast,
        device=device,
        profile=profile,
        att_name=attention_name,
    )


def instantiate_xformer(
    activation: Activation,
    attention_name: str,
    attn_dropout: float,
    causal: bool,
    feedforward_name: str,
    heads: int,
    residual_dropout: float,
    sequence_length: int,
    embed_dim: int,
    dropout: float,
) -> xFormerEncoderBlock:

    block_size = 16

    attention_config = {
        "name": attention_name,
        "dropout": attn_dropout,
        "causal": causal,
        "seq_len": sequence_length,
        "attention_query_mask": _get_attention_query_mask(
            sequence_length, _GLOBAL_ATTENTION_RATIO
        ),
        "num_heads": heads,
        "dim_head": embed_dim / heads,
        "layout": torch.eye(
            sequence_length // block_size,
            sequence_length // block_size,
            dtype=torch.long,
        )
        .unsqueeze(0)
        .expand(heads, -1, -1),
        "block_size": block_size,
    }

    multi_head_config = {
        "num_heads": heads,
        "dim_model": embed_dim,
        "residual_dropout": residual_dropout,
        "attention": attention_config,
    }

    feedforward_config = {
        "name": feedforward_name,
        "dim_model": embed_dim,
        "dropout": dropout,
        "activation": activation,
        "hidden_layer_multiplier": 4,
    }

    position_embedding_config = {
        "name": "sine",
        "dim_model": embed_dim,
        "seq_len": sequence_length,
    }

    block_config = xFormerEncoderConfig(
        dim_model=embed_dim,
        multi_head_config=multi_head_config,
        feedforward_config=feedforward_config,
        position_encoding_config=position_embedding_config,
    )

    block = xFormerEncoderBlock.from_config(block_config)
    return block


def plot(args, results: List[Dict[str, Any]]):
    df = pd.DataFrame(results)
    HEADS = args.heads[-1]
    AMP = args.pytorch_amp[-1]
    EMB = args.embedding_dim[-1]
    CAUSAL = args.causal[-1]
    BATCH_SIZE = args.batch_size[-1]
    ACTIVATION = args.activations[-1]

    df_filtered = df[
        (df["activation"] == ACTIVATION)
        & (df["heads"] == HEADS)
        & (df["autocast"] == AMP)
        & (df["embed_dim"] == EMB)
        & (df["causal"] == CAUSAL)
        & (df["batch_size"] == BATCH_SIZE)
    ]

    df_filtered.sort_values(
        by=["sequence_length", "max_memory"], ascending=[False, True], inplace=True
    )
    sns.barplot(
        x="sequence_length",
        y="max_memory",
        hue="attention_name",
        data=df_filtered,
        palette="Set2",
    )
    plt.xlabel("Sequence length")
    plt.ylabel("Max memory being used")
    plt.title("Memory use")
    plt.savefig("memory_vs_attention.png")
    plt.clf()

    df_filtered.sort_values(
        by=["sequence_length", "run_time"], ascending=[False, True], inplace=True
    )
    sns.barplot(
        x="sequence_length",
        y="run_time",
        hue="attention_name",
        data=df_filtered,
        palette="Set2",
    )
    plt.xlabel("Sequence length")
    plt.ylabel("Average epoch time")
    plt.title("Runtime")
    plt.savefig("runtime_vs_attention.png")


if __name__ == "__main__":
    # Get the user requests
    parser = argparse.ArgumentParser(
        "Benchmark different attention mechanisms on various sequence lengths"
    )
    parser.add_argument(
        "-a", "--attentions", nargs="+", default=list(ATTENTION_REGISTRY.keys())
    )
    parser.add_argument("-mlp", "--mlp", nargs="+", default=["MLP"])
    parser.add_argument(
        "-act", "--activations", nargs="+", default=[a.value for a in Activation]
    )
    parser.add_argument(
        "-emb", "--embedding_dim", nargs="+", default=[64, 128, 256], type=int
    )
    parser.add_argument(
        "-sl", "--sequence_length", nargs="+", default=[576, 1024], type=int
    )
    parser.add_argument("-bs", "--batch_size", nargs="+", default=[8, 16, 32], type=int)
    parser.add_argument("-heads", "--heads", nargs="+", default=[8, 16], type=int)

    parser.add_argument("-fp16", "--pytorch_amp", nargs="+", default=[True], type=bool)
    parser.add_argument("-causal", "--causal", nargs="+", default=[False], type=bool)
    parser.add_argument("-plot", "--plot", action="store_true", default=False)
    parser.add_argument(
        "-profile",
        "--profile",
        help="Pofile the runtime and memory",
        action="store_true",
        default=False,
    )

    args = parser.parse_args()

    # Setup the test configs
    constants = {
        "device": torch.device("cuda") if _use_cuda else torch.device("cpu"),
        "num_warmup": 5,
        "num_steps": 10,
        "dropout": 0.1,
        "attn_dropout": 0.1,
        "residual_dropout": 0.1,
        "profile": args.profile,
    }

    param_grid = {
        "autocast": args.pytorch_amp,
        "causal": args.causal,
        "heads": args.heads,
        "activation": args.activations,
        "attention_name": args.attentions,
        "feedforward_name": args.mlp,
        "sequence_length": args.sequence_length,
        "embed_dim": args.embedding_dim,
        "batch_size": args.batch_size,
    }

    print(
        "Testing the following parameters: \n",
        json.dumps(param_grid, sort_keys=True, indent=4),
    )

    grid = ParameterGrid(param_grid)

    grid_outputs = []

    for params in tqdm(grid, total=len(grid)):
        outputs = test_xformer_encoder_block(**constants, **params)  # type: ignore
        results = {**outputs, **params}
        grid_outputs.append(results)

    print(json.dumps(grid_outputs, sort_keys=True, indent=4))

    # Optional plots
    if args.plot:
        plot(args, grid_outputs)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import logging
from dataclasses import dataclass
from enum import Enum
from typing import Optional

import torch
import torch.autograd.profiler as profiler
import torch.nn as nn
import torch.nn.functional as Fn

from xformers.components.attention import Attention, AttentionConfig, register_attention
from xformers.components.attention.core import (
    scaled_dot_product_attention,
    scaled_query_key_softmax,
)


class LandmarkSelection(str, Enum):
    Orthogonal = "orthogonal"
    KMeans = "kmeans"
    KMeans_Spherical = "kmeans_spherical"
    Random = "random"


@dataclass
class OrthoformerAttentionConfig(AttentionConfig):
    """
    num_landmarks           Number of landmarks to use for softmax approximation.
    subsample_fraction      Percentage of q_samples matrix to sample per iteration
    landmark_selection      Landmark selection strategy
    """

    num_landmarks: Optional[int]
    subsample_fraction: Optional[float]
    landmark_selection: Optional[LandmarkSelection]


@register_attention("orthoformer", OrthoformerAttentionConfig)
class OrthoFormerAttention(Attention):
    def __init__(
        self,
        dropout: float,
        num_landmarks: int = 32,
        subsample_fraction: float = 1.0,
        landmark_selection: LandmarkSelection = LandmarkSelection.Orthogonal,
        *args,
        **kwargs,
    ):
        """
        Orthoformer_ attention mechanism.
        ::

            "Keeping Your Eye on the Ball: Trajectory Attention in Video Transformers"
            Patrick, M., Campbell, D., Asano, Y., Misra, I., Metze, F., Feichtenhofer,
            C., Vedaldi, A., Henriques, J. (2021)

            Reference codebase: https://github.com/facebookresearch/Motionformer

        .. _Orthoformer: https://arxiv.org/abs/2106.05392

        """
        super().__init__()

        self.num_landmarks = num_landmarks
        self.attn_drop = nn.Dropout(dropout)
        self.subsample_fraction = subsample_fraction
        self.landmark_selection = landmark_selection

        # Properties specific to this attention mechanism
        self.supports_attention_mask = True
        self.supports_key_padding_mask = False

    def forward(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        att_mask: Optional[torch.Tensor] = None,
        *args,
        **kwargs,
    ):
        N = k.shape[1]

        if self.num_landmarks == N:
            #  Default attention
            x = scaled_dot_product_attention(q, k, v, att_mask)
        else:
            with torch.no_grad(), profiler.record_function("select landmarks"):
                if self.landmark_selection == LandmarkSelection.Orthogonal:
                    landmarks = self._compute_orthogonal_landmarks(q)
                elif self.landmark_selection == LandmarkSelection.Random:
                    half_L = self.num_landmarks // 2
                    landmarks_q = q[:, torch.randint(q.size(1), (half_L,)), :]
                    landmarks_k = k[:, torch.randint(k.size(1), (half_L,)), :]
                    landmarks = torch.cat((landmarks_q, landmarks_k), dim=-2)
                elif self.landmark_selection == LandmarkSelection.KMeans:
                    landmarks = self._cluster_landmarks(q)
                elif self.landmark_selection == LandmarkSelection.KMeans_Spherical:
                    landmarks = self._cluster_landmarks(q, spherical=True)

            if att_mask is not None:
                logging.warning(
                    "Orthoformer: attention mask passed alongside with using landmarks to reduce dimensions. \
                    The two are typically not compatible"
                )
                # FIXME: Should we still accept a mask in that case ?
                att_mask = None

            # pyre-ignore[61]: TODO(T103337542): `landmarks` mistakenly seems
            # like it could be uninitialized.
            kernel_1 = scaled_query_key_softmax(q, landmarks, att_mask)
            # pyre-ignore[61]: TODO(T103337542): `landmarks` mistakenly seems
            # like it could be uninitialized.
            kernel_2 = scaled_query_key_softmax(landmarks, k, att_mask)
            x = torch.matmul(kernel_1, torch.matmul(kernel_2, v))
        x = self.attn_drop(x)
        return x

    def _cluster_landmarks(
        self,
        q: torch.Tensor,
        spherical: bool = False,
        num_iters: int = 6,
    ) -> torch.Tensor:
        """
        Construct set of landmarks by recursively selecting new landmarks
        that are maximally orthogonal to the existing set.
        Returns near orthogonal landmarks with shape (B, M, D).
        """

        num_landmarks = min(self.num_landmarks, q.shape[1])

        if self.subsample_fraction < 1.0:
            num_samples = max(
                int(self.subsample_fraction * q.size(-2)), num_landmarks
            )  # Need at least M/2 samples of queries and keys
            q_samples = q[:, torch.randint(q.size(-2), (num_samples,)), :]  # (B, N, D)
        else:
            q_samples = q  # (B, N, D)

        if spherical:
            q_samples_normalized = Fn.normalize(
                q_samples, p=2, dim=-1
            )  # may need to change default eps to eps=1e-8 for mixed precision compatibility
            landmarks = self._kmeans_spherical(
                q_samples_normalized, num_landmarks, num_iters
            )
        else:
            landmarks = self._kmeans(q_samples, num_landmarks, num_iters)
        return landmarks  # (B, M, D)

    def _kmeans(self, x: torch.Tensor, K: int, num_iters: int = 10):
        """
        Arguments:
            x: (B, N, D)
            K: number of clusters
            num_iters: the number of kmeans updates
        """

        B, N, D = x.size()
        assert K <= N, f"{K} > {N}"

        c = x[
            :, torch.randperm(N, device=x.device)[:K], :
        ].clone()  # initialisation for the centroids

        with profiler.record_function("kmeans"):
            x_i = x.view(B, N, 1, D)
            c_j = c.view(B, 1, K, D)
            counts = c.new_zeros(B, K)
            ones = x.new_ones((B, N))

            for _ in range(num_iters):
                # E step: assign points to the nearest cluster
                D_ij = ((x_i - c_j) ** 2).sum(-1)  # (B, N, K) squared distances
                cl = D_ij.argmin(
                    dim=-1, keepdim=True
                ).long()  # (B, N, 1) index of point to nearest cluster

                # M step: update the centroids
                c.zero_()
                c.scatter_add_(-2, cl.repeat(1, 1, D), x)  # sum of points per cluster
                counts.fill_(1e-6)  # avoid div0
                counts.scatter_add_(
                    -1, cl.squeeze(-1), ones
                )  # number of points per cluster
                c.divide_(counts.unsqueeze(-1))  # compute the average

        return c

    def _kmeans_spherical(self, x: torch.Tensor, K: int, num_iters=10):
        """
        Arguments:
            x: (B, N, D)
        """
        B, N, D = x.size()
        assert K <= N, f"{K} > {N}"

        # initialisation for the centroids
        c = x[:, torch.randperm(N, device=x.device)[:K], :].clone()

        with profiler.record_function("kmeans_spherical"):
            counts = c.new_zeros(B, K)
            ones = x.new_ones((B, N))

            for _ in range(num_iters):
                # E step: assign points to the nearest cluster
                D_ij = torch.matmul(
                    x, c.transpose(-2, -1)
                )  # (B, N, K) cosine similarity
                cl = D_ij.argmax(
                    dim=-1, keepdim=True
                ).long()  # (B, N, 1) index of point to nearest cluster

                # M step: update the centroids
                c.zero_()
                c.scatter_add_(-2, cl.repeat(1, 1, D), x)  # sum of points per cluster
                counts.fill_(1e-6)  # avoid div0
                counts.scatter_add_(
                    -1, cl.squeeze(-1), ones
                )  # number of points per cluster
                c.divide_(counts.unsqueeze(-1))  # compute the average
                c = Fn.normalize(c, p=2, dim=-1)  # renormalise
        return c

    def _compute_orthogonal_landmarks(self, q: torch.Tensor) -> torch.Tensor:
        """
        Construct set of landmarks by recursively selecting new landmarks
        that are maximally orthogonal to the existing set.
        Returns near orthogonal landmarks with shape (B, M, D).
        """

        if self.subsample_fraction < 1.0:
            # Need at least M samples of queries
            num_samples = max(
                int(self.subsample_fraction * q.size(-2)), self.num_landmarks
            )
            q_samples = q[
                :, torch.randint(q.size(-2), (num_samples,), device=q.device), :
            ]
        else:
            # (B, N, D)
            q_samples = q

        # may need to change default eps to eps=1e-8 for mixed precision compatibility
        q_samples_normalized = Fn.normalize(q_samples, p=2, dim=-1)
        B, N, D = q_samples_normalized.shape

        selected_mask = torch.zeros((B, N, 1), device=q_samples_normalized.device)
        landmark_mask = torch.ones(
            (B, 1, 1), dtype=selected_mask.dtype, device=q_samples_normalized.device
        )

        #  Get initial random landmark
        random_idx = torch.randint(
            q_samples_normalized.size(-2), (B, 1, 1), device=q_samples_normalized.device
        )
        selected_mask.scatter_(-2, random_idx, landmark_mask)

        #  Selected landmarks
        selected_landmarks = torch.empty(
            (B, self.num_landmarks, D),
            device=q_samples_normalized.device,
            dtype=q_samples_normalized.dtype,
        )
        selected_landmarks[:, 0, :] = q_samples_normalized[
            torch.arange(q_samples_normalized.size(0)), random_idx.view(-1), :
        ].view(B, D)

        # Store computed cosine similarities
        cos_sims = torch.empty(
            (B, N, self.num_landmarks),
            device=q_samples_normalized.device,
            dtype=q_samples_normalized.dtype,
        )

        for M in range(1, self.num_landmarks):
            with profiler.record_function("find new landmark"):
                #  Calculate absolute cosine similarity between selected and unselected landmarks
                # (B, N, D) * (B, D) -> (B, N)
                cos_sims[:, :, M - 1] = torch.einsum(
                    "b n d, b d -> b n",
                    q_samples_normalized,
                    selected_landmarks[:, M - 1, :],
                ).abs()

                # (B, N, M) cosine similarities of current set of landmarks wrt all queries and keys
                cos_sim_set = cos_sims[:, :, :M]

                #  Get orthogonal landmark: landmark with smallest absolute cosine similarity:
                # set cosine similarity for already selected landmarks to > 1
                cos_sim_set.view(-1, M)[selected_mask.flatten().bool(), :] = 10

                # (B,) - want max for non
                selected_landmark_idx = cos_sim_set.amax(-1).argmin(-1)

                #  Add most orthogonal landmark to selected landmarks:
                selected_landmarks[:, M, :] = q_samples_normalized[
                    torch.arange(q_samples_normalized.size(0)), selected_landmark_idx, :
                ].view(B, D)

                #  Removed selected indices from non-selected mask:
                selected_mask.scatter_(
                    -2, selected_landmark_idx.unsqueeze(-1).unsqueeze(-1), landmark_mask
                )

        # (B, M, D)
        landmarks = torch.masked_select(q_samples, selected_mask.bool()).reshape(
            B, -1, D
        )
        return landmarks  #  (B, M, D)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import logging
from dataclasses import asdict
from typing import Optional, Tuple, Union

import torch
import torch.nn as nn

from xformers.components import (
    PatchEmbeddingConfig,
    PostNorm,
    PreNorm,
    Residual,
    ResidualNormStyle,
    build_multi_head_attention,
    build_patch_embedding,
)
from xformers.components.feedforward import build_feedforward
from xformers.components.positional_embedding import build_positional_embedding
from xformers.components.residual import get_deepnorm_coefficients
from xformers.components.simplicial_embedding import SimplicialEmbedding
from xformers.factory.block_configs import (
    NormalizationType,
    xFormerDecoderConfig,
    xFormerEncoderConfig,
)


def _get_ln_factory(
    d_model: int,
    residual_norm_style: Optional[ResidualNormStyle],
    use_triton: bool,
    residual: bool,
    normalization: NormalizationType = NormalizationType.LayerNorm,
    residual_scale: float = 1.0,
):
    """
    Handle all the supported residual path configurations.

    ..Note: we return the appropriate constructor, not an actual layer
    """

    def get_layer_wrapper(
        d_model: int,
        sublayer: nn.Module,
        residual_norm_style: Optional[ResidualNormStyle],
        residual: bool,
        residual_scale: float,
    ):
        if residual:
            if residual_norm_style == ResidualNormStyle.Pre:
                return Residual(
                    layer=PreNorm(d_model, sublayer, normalization, use_triton),
                    scale=None,
                )
            elif residual_norm_style == ResidualNormStyle.Post:
                return PostNorm(
                    d_model,
                    Residual(layer=sublayer, scale=None),
                    normalization,
                    use_triton,
                )
            elif residual_norm_style == ResidualNormStyle.DeepNorm:
                return PostNorm(
                    d_model,
                    Residual(layer=sublayer, scale=residual_scale),
                    normalization,
                    use_triton=use_triton,
                )
            else:
                raise ValueError

        return (
            PreNorm(d_model, sublayer, normalization, use_triton)
            if residual_norm_style == ResidualNormStyle.Pre
            else PostNorm(d_model, sublayer, normalization, use_triton)
        )

    def ln_factory(sublayer: nn.Module):
        return get_layer_wrapper(
            d_model, sublayer, residual_norm_style, residual, residual_scale
        )

    return ln_factory


class xFormerEncoderBlock(torch.nn.Module):
    r"""A vanilla Transformer Encoder block"""

    def __init__(self, config: xFormerEncoderConfig, **kwargs):
        super().__init__()

        self.reversible_f = None
        self.reversible_g = None
        self.residual_norm_style = config.residual_norm_style
        self.dim_model = config.dim_model

        # If this layer is the first one, and a pose encoding has been requested
        if (
            config.position_encoding_config is not None
            and config.layer_position.is_first()
        ):
            self.pose_encoding = build_positional_embedding(
                asdict(config.position_encoding_config)
            )

            pos_encoding_dim = config.position_encoding_config.dim_model
            mha_dim = config.multi_head_config["dim_model"]

            if pos_encoding_dim != mha_dim:
                logging.warning(
                    f"The embedding dim and model dim do not match ({pos_encoding_dim} vs {mha_dim}), adding a projector layer."  # noqa
                )
                self.embedding_projector = nn.Linear(pos_encoding_dim, mha_dim)
        else:
            self.pose_encoding = None

        if config.residual_norm_style == ResidualNormStyle.DeepNorm:
            # Just use the layer norm coefficient here,
            # the init will be handled at the xformers level (knows about encoder and decoder blocks)
            deep_norm_coefficients, _ = get_deepnorm_coefficients(
                encoder_layers=config.num_layers, decoder_layers=0
            )
            assert deep_norm_coefficients is not None
            residual_scale = deep_norm_coefficients.alpha
        else:
            residual_scale = 1.0

        # mini helper, builds a normalization layer with the right Pre/Post config, residuals, and the right dimensions
        ln_factory = _get_ln_factory(
            config.dim_model,
            config.residual_norm_style,
            use_triton=config.use_triton,
            residual=True,
            residual_scale=residual_scale,
            normalization=config.normalization,
        )

        mha = build_multi_head_attention(config.multi_head_config)
        feedforward = build_feedforward(asdict(config.feedforward_config))

        # Expose attention specific capabilities
        self.supports_attention_mask = mha.attention.supports_attention_mask
        self.requires_same_k_q_dimensions = mha.attention.requires_same_k_q_dimensions
        self.causal = (
            mha.attention.causal if hasattr(mha.attention, "causal") else False
        )

        # Wrappers handle the different layer norm styles (pre- and post-) and the residual path
        self.wrap_att = ln_factory(mha)
        self.wrap_ff: Union[Residual, PostNorm] = ln_factory(feedforward)
        if (
            config.residual_norm_style == ResidualNormStyle.Pre
            and config.layer_position.is_last()
        ):
            self.wrap_ff = PostNorm(
                config.dim_model,
                self.wrap_ff,
                normalization=config.normalization,
                use_triton=config.use_triton,
            )

        # Simplicial embeddings are only used if specified, and on the last layer
        self.simplicial_embedding: Optional[SimplicialEmbedding] = None
        if config.simplicial_embeddings is not None and config.layer_position.is_last():
            self.simplicial_embedding = SimplicialEmbedding(
                **config.simplicial_embeddings
            )

        # Optional patch embedding
        self.patch_emb: Optional[nn.Module] = None

        if config.patch_embedding_config is not None:
            self.patch_emb = build_patch_embedding(
                PatchEmbeddingConfig(**config.patch_embedding_config)
            )

    @classmethod
    def from_config(cls, config: xFormerEncoderConfig):
        return cls(config)

    @staticmethod
    def get_reversible_layer(config) -> Tuple[nn.Module, nn.Module]:
        ln_factory = _get_ln_factory(
            config.dim_model,
            config.residual_norm_style,
            residual=False,
            use_triton=config.use_triton,
            normalization=config.normalization,
        )

        mha = build_multi_head_attention(config.multi_head_config)
        feedforward = build_feedforward(asdict(config.feedforward_config))

        reversible_f = ln_factory(mha)
        reversible_g = ln_factory(feedforward)
        return reversible_f, reversible_g

    def forward(
        self,
        x: torch.Tensor,
        att_mask: Optional[torch.Tensor] = None,
        input_mask: Optional[torch.Tensor] = None,
    ):
        if self.patch_emb is not None:
            x = self.patch_emb(x)

        if self.pose_encoding is not None:
            x = self.pose_encoding(x)

            if hasattr(self, "embedding_projector"):
                x = self.embedding_projector(x)

        # Handle the optional input masking, differs on Q, K, V
        if input_mask is not None:
            q = x
            k = x * input_mask.unsqueeze(-1)
            v = k
        else:
            q, k, v = x, x, x

        # Pre/Post norms and residual paths are already handled
        x = self.wrap_att(inputs=[q, k, v], att_mask=att_mask)
        x = self.wrap_ff(inputs=[x])

        # Optional simplicial embeddings
        if self.simplicial_embedding is not None:
            x = self.simplicial_embedding(x)

        return x


class xFormerDecoderBlock(torch.nn.Module):
    r"""A vanilla Transformer Decoder block

    ... note: this implementation is not (yet ?) reversible"""

    def __init__(self, config: xFormerDecoderConfig, **kwargs):
        super().__init__()

        # If this layer is the first one, and a pose encoding as been requested
        if (
            config.position_encoding_config is not None
            and config.layer_position.is_first()
        ):
            self.pose_encoding = build_positional_embedding(
                config.position_encoding_config
            )

            pos_encoding_dim = config.position_encoding_config.dim_model
            mha_dim = config.multi_head_config_masked["dim_model"]

            if pos_encoding_dim != mha_dim:

                logging.warning(
                    f"The embedding dim and model dim do not match ({pos_encoding_dim} vs {mha_dim}), adding a projector layer."  # noqa
                )

                self.embedding_projector = nn.Linear(pos_encoding_dim, mha_dim)
        else:
            self.pose_encoding = None

        if config.residual_norm_style == ResidualNormStyle.DeepNorm:
            # Just use the layer norm coefficient here,
            # the init will be handled at the xformers level (knows about encoder and decoder blocks)
            _, deep_norm_coefficients = get_deepnorm_coefficients(
                encoder_layers=0, decoder_layers=config.num_layers
            )
            assert deep_norm_coefficients is not None
            residual_scale = deep_norm_coefficients.alpha
        else:
            residual_scale = 1.0

        # mini helper, builds a LayerNorm with the right Pre/Post config and the right dimensions
        ln_factory = _get_ln_factory(
            config.dim_model,
            config.residual_norm_style,
            use_triton=config.use_triton,
            residual=True,
            residual_scale=residual_scale,
            normalization=config.normalization,
        )

        mha = build_multi_head_attention(config.multi_head_config_masked)
        cross_mha = build_multi_head_attention(config.multi_head_config_cross)
        feedforward = build_feedforward(config.feedforward_config)

        # Expose attention or feedforward specific capabilities
        self.supports_attention_mask = mha.attention.supports_attention_mask
        self.requires_same_k_q_dimensions = mha.attention.requires_same_k_q_dimensions
        self.requires_squared_context_length = (
            feedforward.requires_squared_context
            or mha.attention.requires_squared_context
        )

        self.causal_attention = (
            mha.attention.causal if hasattr(mha.attention, "causal") else False
        )

        # Wrappers handle the different layer norm styles (pre- and post-) and the residual path
        self.wrap_att = ln_factory(mha)
        self.wrap_cross = ln_factory(cross_mha)
        self.wrap_ff: Union[Residual, PostNorm] = ln_factory(feedforward)

        if (
            config.residual_norm_style == ResidualNormStyle.Pre
            and config.layer_position.is_last()
        ):
            self.wrap_ff = PostNorm(
                config.dim_model,
                self.wrap_ff,
                normalization=NormalizationType.LayerNorm,
            )

    @classmethod
    def from_config(cls, config: xFormerDecoderConfig):
        return cls(config)

    def forward(
        self,
        target: torch.Tensor,
        memory: torch.Tensor,
        encoder_att_mask: Optional[torch.Tensor] = None,
        decoder_att_mask: Optional[torch.Tensor] = None,
        input_mask: Optional[torch.Tensor] = None,
    ):
        if self.pose_encoding is not None:
            target = self.pose_encoding(target)

            if hasattr(self, "embedding_projector"):
                target = self.embedding_projector(target)

        # Handle the optional input masking, differs on Q, K, V
        if input_mask is not None:
            target_q = target
            target_k = target * input_mask.unsqueeze(-1)
            target_v = target_k
        else:
            target_q, target_k, target_v = target, target, target

        x = self.wrap_att(
            inputs=[target_q, target_k, target_v], att_mask=decoder_att_mask
        )
        x = self.wrap_cross(inputs=[x, memory, memory], att_mask=encoder_att_mask)
        x = self.wrap_ff(inputs=[x])

        return x



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import distutils.command.clean
import glob
import os
import re
import shutil
import subprocess
import sys
from pathlib import Path

import setuptools
import torch
from torch.utils.cpp_extension import (
    CUDA_HOME,
    BuildExtension,
    CppExtension,
    CUDAExtension,
)

this_dir = os.path.dirname(os.path.abspath(__file__))


def fetch_requirements():
    with open("requirements.txt") as f:
        reqs = f.read().strip().split("\n")
    return reqs


# https://packaging.python.org/guides/single-sourcing-package-version/
def find_version(version_file_path):
    with open(version_file_path) as version_file:
        version_match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]", version_file.read(), re.M
        )
        # The following is used to build release packages.
        # Users should never use it.
        suffix = os.getenv("XFORMERS_VERSION_SUFFIX", "")
        if version_match:
            return version_match.group(1) + suffix
        raise RuntimeError("Unable to find version string.")


def get_cuda_version(cuda_dir) -> int:
    nvcc_bin = "nvcc" if cuda_dir is None else cuda_dir + "/bin/nvcc"
    raw_output = subprocess.check_output([nvcc_bin, "-V"], universal_newlines=True)
    output = raw_output.split()
    release_idx = output.index("release") + 1
    release = output[release_idx].split(".")
    bare_metal_major = int(release[0])
    bare_metal_minor = int(release[1][0])

    assert bare_metal_minor < 100
    return bare_metal_major * 100 + bare_metal_minor


def get_flash_attention_extensions(cuda_version: int, extra_compile_args):
    # Figure out default archs to target
    DEFAULT_ARCHS_LIST = ""
    if cuda_version > 1100:
        DEFAULT_ARCHS_LIST = "7.5;8.0;8.6"
    elif cuda_version == 1100:
        DEFAULT_ARCHS_LIST = "7.5;8.0"
    else:
        return []

    if os.getenv("XFORMERS_DISABLE_FLASH_ATTN", "0") != "0":
        return []

    archs_list = os.environ.get("TORCH_CUDA_ARCH_LIST", DEFAULT_ARCHS_LIST)
    nvcc_archs_flags = []
    for arch in archs_list.split(";"):
        assert len(arch) >= 3, f"Invalid sm version: {arch}"

        num = 10 * int(arch[0]) + int(arch[2])
        # Need at least 7.5
        if num < 75:
            continue
        nvcc_archs_flags.append(f"-gencode=arch=compute_{num},code=sm_{num}")
        if arch.endswith("+PTX"):
            nvcc_archs_flags.append(f"-gencode=arch=compute_{num},code=compute_{num}")
    if not nvcc_archs_flags:
        return []

    this_dir = os.path.dirname(os.path.abspath(__file__))
    flash_root = os.path.join(this_dir, "third_party", "flash-attention")
    if not os.path.exists(flash_root):
        raise RuntimeError(
            "flashattention submodule not found. Did you forget "
            "to run `git submodule update --init --recursive` ?"
        )

    return [
        CUDAExtension(
            name="xformers._C_flashattention",
            sources=[
                os.path.join(this_dir, "third_party", "flash-attention", path)
                for path in [
                    "csrc/flash_attn/fmha_api.cpp",
                    "csrc/flash_attn/src/fmha_fprop_fp16_kernel.sm80.cu",
                    "csrc/flash_attn/src/fmha_dgrad_fp16_kernel_loop.sm80.cu",
                    "csrc/flash_attn/src/fmha_block_fprop_fp16_kernel.sm80.cu",
                    "csrc/flash_attn/src/fmha_block_dgrad_fp16_kernel_loop.sm80.cu",
                ]
            ],
            extra_compile_args={
                **extra_compile_args,
                "nvcc": extra_compile_args.get("nvcc", [])
                + [
                    "-O3",
                    "-U__CUDA_NO_HALF_OPERATORS__",
                    "-U__CUDA_NO_HALF_CONVERSIONS__",
                    "--expt-relaxed-constexpr",
                    "--expt-extended-lambda",
                    "--use_fast_math",
                    "--ptxas-options=-v",
                    "-lineinfo",
                ]
                + nvcc_archs_flags,
            },
            include_dirs=[
                Path(flash_root) / "csrc" / "flash_attn",
                Path(flash_root) / "csrc" / "flash_attn" / "src",
                #            Path(flash_root) / 'csrc' / 'flash_attn' / 'cutlass' / 'include',
                Path(this_dir) / "third_party" / "cutlass" / "include",
            ],
        )
    ]


def get_extensions():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    extensions_dir = os.path.join(
        this_dir, "xformers", "components", "attention", "csrc"
    )

    main_file = glob.glob(os.path.join(extensions_dir, "*.cpp"))

    source_cpu = glob.glob(os.path.join(extensions_dir, "cpu", "*.cpp")) + glob.glob(
        os.path.join(extensions_dir, "autograd", "*.cpp")
    )

    sources = main_file + source_cpu

    source_cuda = glob.glob(
        os.path.join(extensions_dir, "cuda", "**", "*.cu"), recursive=True
    )

    sputnik_dir = os.path.join(this_dir, "third_party", "sputnik")
    cutlass_dir = os.path.join(this_dir, "third_party", "cutlass", "include")
    if not os.path.exists(cutlass_dir):
        raise RuntimeError(
            "CUTLASS submodule not found. Did you forget "
            "to run `git submodule update --init --recursive` ?"
        )

    extension = CppExtension

    define_macros = []

    extra_compile_args = {"cxx": ["-O3"]}
    if sys.platform == "win32":
        define_macros += [("xformers_EXPORTS", None)]
        extra_compile_args["cxx"].append("/MP")
    elif "OpenMP not found" not in torch.__config__.parallel_info():
        extra_compile_args["cxx"].append("-fopenmp")

    include_dirs = [extensions_dir]
    ext_modules = []

    if (torch.cuda.is_available() and ((CUDA_HOME is not None))) or os.getenv(
        "FORCE_CUDA", "0"
    ) == "1":
        extension = CUDAExtension
        sources += source_cuda
        include_dirs += [sputnik_dir, cutlass_dir]
        nvcc_flags = os.getenv("NVCC_FLAGS", "")
        if nvcc_flags == "":
            nvcc_flags = ["--use_fast_math", "-DNDEBUG"]
        else:
            nvcc_flags = nvcc_flags.split(" ")
        cuda_version = get_cuda_version(CUDA_HOME)
        if cuda_version >= 1102:
            nvcc_flags += [
                "--threads",
                "4",
                "--ptxas-options=-v",
            ]
        extra_compile_args["nvcc"] = nvcc_flags

        ext_modules += get_flash_attention_extensions(
            cuda_version=cuda_version, extra_compile_args=extra_compile_args
        )

    sources = [os.path.join(extensions_dir, s) for s in sources]

    ext_modules.append(
        extension(
            "xformers._C",
            sorted(sources),
            include_dirs=include_dirs,
            define_macros=define_macros,
            extra_compile_args=extra_compile_args,
        )
    )

    return ext_modules


class clean(distutils.command.clean.clean):  # type: ignore
    def run(self):
        if os.path.exists(".gitignore"):
            with open(".gitignore", "r") as f:
                ignores = f.read()
                for wildcard in filter(None, ignores.split("\n")):
                    for filename in glob.glob(wildcard):
                        try:
                            os.remove(filename)
                        except OSError:
                            shutil.rmtree(filename, ignore_errors=True)

        # It's an old-style class in Python 2.7...
        distutils.command.clean.clean.run(self)


if __name__ == "__main__":
    setuptools.setup(
        name="xformers",
        description="XFormers: A collection of composable Transformer building blocks.",
        version=find_version(os.path.join(this_dir, "xformers", "__init__.py")),
        setup_requires=[],
        install_requires=fetch_requirements(),
        packages=setuptools.find_packages(exclude=("tests", "tests.*")),
        ext_modules=get_extensions(),
        cmdclass={
            "build_ext": BuildExtension.with_options(no_python_abi_suffix=True),
            "clean": clean,
        },
        url="https://facebookresearch.github.io/xformers/",
        python_requires=">=3.6",
        author="Facebook AI Research",
        author_email="lefaudeux@fb.com",
        long_description="XFormers: A collection of composable Transformer building blocks."
        + "XFormers aims at being able to reproduce most architectures in the Transformer-family SOTA,"
        + "defined as compatible and combined building blocks as opposed to monolithic models",
        long_description_content_type="text/markdown",
        classifiers=[
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "License :: OSI Approved :: BSD License",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Operating System :: OS Independent",
        ],
        zip_safe=False,
    )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


# CREDITS: This comes almost as-is from the Triton layer norm tutorial
# https://github.com/openai/triton/blob/master/python/tutorials/05-layer-norm.py


import triton
import triton.language as tl


# fmt: off
@triton.jit
def layer_norm_fw(X, Y, W, B, M, V, stride, N, eps, affine: tl.constexpr, BLOCK_SIZE_N: tl.constexpr):
    # fmt: on
    """
    Fused layernorm kernel over a 3d tensor.
    The layer norm is applied over the last dimension.

    Compute
        y = (x - E(x))/(sqrt(var(x) + epsilon)) * gamma + beta
    """

    row = tl.program_id(0)
    cols = tl.arange(0, BLOCK_SIZE_N)
    mask = cols < N

    # Move to this row
    x_ptrs = X + row * stride + cols
    x = tl.load(x_ptrs, mask=mask, other=0.0).to(tl.float32)

    # Compute mean and variance
    mean = tl.sum(x, axis=0) / N
    x_zm = tl.where(mask, x - mean, 0.0)
    tl.store(M + row, mean)

    x_var = tl.sum(x_zm * x_zm, axis=0) / N
    rstd = 1.0 / tl.sqrt(x_var + eps)

    # Normalize, optionally affine
    y = x_zm * rstd
    tl.store(V + row, rstd)

    mask = cols < N
    if affine:
        w = tl.load(W + cols, mask=mask, other=1.0)
        b = tl.load(B + cols, mask=mask, other=0.0)
        y = y * w + b

    y_ptrs = Y + row * stride + cols
    tl.store(y_ptrs, y, mask=mask)


# Backward pass (DX + partial DW + partial DB)
# fmt: off
@triton.jit
def layer_norm_bwd_dx_fused(
    DX, DY, DW, DB,
    X, W, M, V,
    Lock, stride, N,
    # META-parameters
    affine: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
):
    # fmt: on

    # position of elements processed by this program
    row = tl.program_id(0)
    cols = tl.arange(0, BLOCK_SIZE_N)
    mask = cols < N

    # offset data pointers to start at the row of interest
    x_ptrs = X + row * stride + cols
    dy_ptrs = DY + row * stride + cols

    # load data to SRAM
    x = tl.load(x_ptrs, mask=mask, other=0)
    dy = tl.load(dy_ptrs, mask=mask, other=0)
    mean = tl.load(M + row)
    rstd = tl.load(V + row)

    # compute dx
    xhat = (x - mean) * rstd

    if affine:
        w = tl.load(W + cols, mask=mask, other=0)
        wdy = w * dy
    else:
        wdy = dy

    xhat = tl.where(mask, xhat, 0.)
    wdy = tl.where(mask, wdy, 0.)
    mean1 = tl.sum(xhat * wdy, axis=0) / N
    mean2 = tl.sum(wdy, axis=0) / N
    dx = (wdy - (xhat * mean1 + mean2)) * rstd

    # write-back dx
    cols = tl.arange(0, BLOCK_SIZE_N)
    mask = cols < N  # re-materialize the mask to save registers
    dx_ptrs = DX + row * stride + cols
    tl.store(dx_ptrs, dx, mask=mask)

    if affine:
        # accumulate partial sums for dw/db
        partial_dw = (dy * xhat).to(w.dtype)
        partial_db = dy.to(w.dtype)

        # offset locks and weight/bias gradient pointer
        # each kernel instance accumulates partial sums for
        # DW and DB into one of GROUP_SIZE_M independent buffers
        # these buffers stay in the L2, which allow this kernel
        # to be fast
        lock_id = row % GROUP_SIZE_M
        Lock += lock_id
        Count = Lock + GROUP_SIZE_M

        # - wait for a lock on the accumulated dw/db
        while tl.atomic_cas(Lock, 0, 1) == 1:
            pass
        count = tl.load(Count)

        # - we got the lock, accumulate this kernel's results with
        # the stored values.
        dw_ptrs = DW + lock_id * N + cols
        db_ptrs = DB + lock_id * N + cols

        if count == 0:
            # first store doesn't accumulate
            tl.atomic_xchg(Count, 1)
        else:
            partial_dw += tl.load(dw_ptrs, mask=mask, other=0.)
            partial_db += tl.load(db_ptrs, mask=mask, other=0.)

        tl.store(dw_ptrs, partial_dw, mask=mask)
        tl.store(db_ptrs, partial_db, mask=mask)

        # release lock
        tl.atomic_xchg(Lock, 0)


# Backward pass (total DW + total DB)
# fmt: off
@triton.jit
def layer_norm_bwd_dwdb(
    DW, DB, FINAL_DW, FINAL_DB,
    M, N,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr
):
    # fmt: on

    pid = tl.program_id(0)

    cols = pid * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    mask_cols = cols < N

    dw = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    db = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)

    for i in range(0, M, BLOCK_SIZE_M):
        rows = i + tl.arange(0, BLOCK_SIZE_M)
        offs = rows[:, None] * N + cols[None, :]
        mask_rm = rows < M

        dw += tl.load(DW + offs, mask=mask_rm[:, None] & mask_cols[None, :], other=0.0)
        db += tl.load(DB + offs, mask=mask_rm[:, None] & mask_cols[None, :], other=0.0)

    sum_dw = tl.sum(dw, axis=0)
    sum_db = tl.sum(db, axis=0)

    cols = pid * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    mask_cols = cols < N

    tl.store(FINAL_DW + cols, sum_dw, mask=mask_cols)
    tl.store(FINAL_DB + cols, sum_db, mask=mask_cols)
